{- | Back-end translation of FOL plugin: generates .otter syntax from 
Formulas optained by "Plugins.FOL.Translation". -}

module Plugins.FOL.Otter where

import Control.Monad.State
import Data.Char	    ( isAlpha, toUpper, toLower )
import Text.PrettyPrint

import Utils.Monads
import Plugins.FOL.Types

-- Otter printing ---------------------------------------------------------

{- | Once we have formulas, we translate them to clauses which we can 
     printout to the first-order prover
-}

data Lit = LNeg String [FOLTerm]
	 | LPos String [FOLTerm]

data FOLClause = FOLClause [Lit]

{- | Translate a list a FOL formulas which have been generated from an
agdaLight type into a list of clauses for output into a @.otter@ file
-}
translateFormulas :: [Formula] -> [[FOLClause]]
translateFormulas formulas =
    evalState (mapM (\f -> map FOLClause <$> clausify [] f) formulas) 0
{- U: here is not a bug anymore.
   The new skolem constants is no longer locally named
   for each formula, but for *all* formulas.
   Otherwise, the list of formulas

     Exists x. x == 0
     Exists x. x == 1

   would be translated as

     x0 == 0
     x0 == 1,
 
   hence, 0 == 1. The correct translation, of course, is

     x0 == 0
     x1 == 1.

  Currently, this does not affect us, since we work in propositional logic 
  only.
-}

tick :: State Int Int
tick =	do  n <- get
	    put (n + 1)
	    return n

{- | The list of 'FOLTerm' is just a list of 'FOLVar'. 
The new skolem constants introduced for 'Exists' can depend on all 'FOLVar'
introduced before.
The output of 'clausify' is a CNF. -}
clausify :: [FOLTerm] -> Formula -> State Int [[Lit]] 
clausify vs p = 
  case positivize p of                 
    ForAll x f -> 
        do id <- tick
           let v = FOLVar (x ++ show id)
           clausify (v:vs) (f v)
    Exists x f -> 
        do id <- tick
           let skol = FOLFun (x ++ show id) 
           clausify vs (f (skol vs))  -- only use of vs
    And p1 p2 -> 
        do xs <- clausify vs p1
           ys <- clausify vs p2
           return (xs++ys)
    Or p1 p2 -> 
        do xs <- clausify vs p1
           ys <- clausify vs p2
           return (xs `cross` ys)
        where xs `cross` ys = [ x ++ y | x <- xs, y <- ys ]     
    Not (Predicate p el) -> return [[LNeg p el]]
    Predicate p el -> return [[LPos p el]]
    TRUE	    -> return []
    FALSE	    -> return [[LNeg "true" []]]
    -- the case distinction is complete 
    -- because of the shape of positive formulas

{- | 'positivize' creates a positive formula, which is either a
positive or negative literal or some formula not starting with @Not@.  
It also resolves @Implies@ and @Equiv@. -}
positivize :: Formula -> Formula
positivize p = case p of 
    Not p           -> negativize p
    Implies p1 p2   -> impliesDef p1 p2 -- positive formula
    Equiv p1 p2     -> equivDef p1 p2   -- positive formula
    p               -> p

{- | 'negativize' creates a positive formula with the opposite meaning. -}
negativize :: Formula -> Formula
negativize p = case p of 
    Implies p1 p2   -> negativize $ impliesDef p1 p2
    Equiv p1 p2     -> negativize $ equivDef p1 p2
    ForAll x f      -> Exists x (nt . f) 
    Exists x f      -> ForAll x (nt . f)
    And p1 p2       -> Or (nt p1) (nt p2) 
    Or p1 p2        -> And (nt p1) (nt p2) 
    Not p           -> positivize p
    TRUE	    -> FALSE
    FALSE	    -> TRUE
    p               -> Not p  -- atom

impliesDef p q = Or (nt p) q
equivDef p q = And (impliesDef p q) (impliesDef q p)

{- | 'nt' is just a tiny optimization, not needed, since positivize would 
take care of double negation.
A: I am not sure this optimization buys one anything. -}
nt :: Formula -> Formula
nt (Not p) = p
nt p       = Not p

-- ? * Pretty-printing functions 

-- ? ** Otter Format 

ppConst :: String -> Doc
ppConst s = text $ 'c':s
{-
ppConst (c:cs) | isAlpha c = text $ toLower c : cs
               | otherwise = text $ "c" ++ c : cs
-}

ppFOLVar :: String -> Doc
ppFOLVar s = text $ 'X':s
{-
ppFOLVar (c:cs)| isAlpha c = text $ toUpper c : cs
               | otherwise = text $ "X" ++ c : cs
-}
ppTerm :: FOLTerm -> Doc
ppTerm t = 
    case t of 
    FOLFun s [] -> ppConst (deOp s)
    FOLFun s l	-> ppConst (deOp s) <> ppTerms l
    FOLVar s	-> ppFOLVar (deOp s)

ppTerms :: [FOLTerm] -> Doc
ppTerms l = parens (hsep (punctuate comma (map ppTerm l)))

ppLit :: Lit -> Doc
ppLit l = 
    case l of
    LNeg s tl -> (text "-") <> ppBody s tl
    LPos s tl -> ppBody s tl
    where ppBody s [] = text (deOp s)
	  ppBody s tl = text (deOp s) <> ppTerms tl

-- Is this necessary? According to Koen, Yes.
deOp :: String -> String
deOp s@(c:_)
    | isAlpha c	= s
    | otherwise	= "op__" ++ concatMap deSym s
    where
	deSym '!'   = "bang"
	deSym '#'   = "hash"
	deSym '$'   = "dollar"
	deSym '%'   = "percent"
	deSym '&'   = "amp"
	deSym '*'   = "times"
	deSym '+'   = "plus"
	deSym '.'   = "dot"
	deSym '/'   = "slash"
	deSym '<'   = "less"
	deSym '='   = "eq"
	deSym '>'   = "greater"
	deSym '?'   = "question"
	deSym '@'   = "at"
	deSym '\\'  = "backslash"
	deSym '^'   = "hat"
	deSym '|'   = "bar"
	deSym '-'   = "minus"
	deSym '~'   = "twiddle"
	deSym ':'   = "colon"
	deSym c	    = [c]

ppFOLClause :: FOLClause -> Doc
ppFOLClause (FOLClause l) = (hsep (punctuate (text " |") (map ppLit l))) 
                      <> (text ".")

makeOtterFile :: ([FOLClause],[FOLClause]) -> String
makeOtterFile (cs,p) = render complete_doc
    where complete_doc = header $+$ transd $+$ 
                         middle $+$ transp $+$ footer
          transd = vcat $ map ppFOLClause cs
          transp = vcat $ map ppFOLClause p
          header = vcat $ map text [ "set(prolog_style_variables).",
                                     "set(auto).",
                                     "",
                                     "list(usable).",
                                     "",
                                     "equal(X,X).",
				     "true.",
                                     "" ]
          middle = vcat $ map text [ "",
				     "end_of_list.",
                                     "",
                                     "list(sos).",
                                     "" ]
          footer = vcat $ map text [ "",
                                     "end_of_list.",
                                     "" ]

