{-# OPTIONS -fglasgow-exts #-}

module Plugins.FOL.Translation where

import Control.Monad.RWS
import Control.Monad.Error
import Data.Set ( Set, emptySet, elementOf, addToSet, setToList, mkSet )

import Syntax.Internal

import TypeChecker
import TypeChecker.Monad
import TypeChecker.Monad.Errors
import TypeChecker.Monad.Definitions
import TypeChecker.Monad.Options
import TypeChecker.Values   ( vApp, valueType )
import TypeChecker.Equality
import TypeChecker.UnfoldConstants
import TypeChecker.Computation
import TypeChecker.Values
import TypeChecker.Terms

import Utils.Error
import Utils.Map
import Utils.Monads
import Utils.Names		( fresh )
--import Utils.Debug

import Plugins.FOL.Types
import Plugins.FOL.Constants
import Plugins.FOL.Errors
import Plugins.FOL.Computation	( unfoldAllButFOLConstants )
import Plugins.FOL.Substitution

{--------------------------------------------------------------------------
    The Translation Monad
 --------------------------------------------------------------------------}

type M = RWST (Set Name) [Formula] Int TC

runM :: M a -> TC (a, [Formula])
runM m = evalRWST m emptySet 0

isBoundVar :: Name -> M Bool
isBoundVar x = elementOf x <$> ask

bindVar :: Name -> M a -> M a
bindVar x = local $ \s -> addToSet s x

{--------------------------------------------------------------------------
    Lambda Lifting
 --------------------------------------------------------------------------}

lambdaName :: M String
lambdaName =
    do	n <- (get :: M Int)
	put (n + 1)
	return $ "lam_" ++ show n

lambdaLift :: Type -> Closure -> M FOLTerm
lambdaLift v f =
    do	name	<- lambdaName
	xs	<- setToList <$> ask
	x	<- lift $ fresh (clName f)
	let call = FOLFun name [] `apps` map FOLVar xs
	    lhs  = app [call, FOLVar x]
	    args = xs ++ [x]
	rhs	<- bindVar x $ valueToTerm (f `clApp` VId x v)
	tell [forAll (length args) $ \ts -> substF args ts $ equal [lhs, rhs]]
	return call
    where
	apps = foldl (\x y -> app [x,y])

{--------------------------------------------------------------------------
    Translation
 --------------------------------------------------------------------------}

{- |
    The top level function. Returns both the value of its argument and the
    corresponding first order formula.
    <p>
    The argument can be one of four things:
	* a primitive constant (e.g. __fol_extensionality)
	* proof of a proposition
	* the name of a datatype
	* the name of a function defined by pattern matching
-}
termToFOL :: Term -> M (Value, Formula)
termToFOL (Var x _)
    | Just p <- lookup x fol_primitives	=
	return	( VId x (error $ "fol primitive: " ++ x), p )
termToFOL e =
    do	e' <- lift $ inferType e
	v  <- lift $ evaluateTypedTerm e'
	(,) v <$> prop e (valueType v) v
    where
	prop e t v
	    | isProp t	= valueToFOL t
	    | otherwise	= lift $ failure $ notOfTypePropError t (valueType t)

substF xs ts f = foldr (uncurry substFormula) f $ zip xs ts

forAll :: Int -> ([FOLTerm] -> Formula) -> Formula
forAll 0 f = f []
forAll n f =	ForAll "x"	$ \x ->
		forAll (n - 1)	$ \xs -> f (x:xs)

valueToFOL :: Value -> M Formula
valueToFOL v =
--    lift $ sayWhen (whenTranslatingToFOL v) $
	do  defs <- lift getDefinitions
	    v' `seq` propToFOL v'
    where
	v' = unfoldAllButFOLConstants v

-- | Should perhaps check if the value is a subtype of Prop. But it doesn't.
isProp :: Value -> Bool
isProp v =
    case unfoldConstants $ valueType v of
	VSort VSProp	-> True
	_		-> False

isSet :: Value -> TC ()
isSet v = valueType v =<= VSort (VSUniv 0)

piToFOL :: Name -> Type -> (Value -> Value) -> (Value -> M Formula) -> M Formula
piToFOL "_" t f toFOL
    | isProp t	= implies <$> toFOL t <*> toFOL (f (VId "_" t))
	    -- We translate independent functions to implications. Sound?
    | otherwise	= toFOL (f (VId "_" t))

piToFOL name t f toFOL =
    do  lift $ sayWhat (higherOrderQuantification t) $ isSet t
	x <- lift $ fresh name
	fm <- bindVar x $ toFOL $ f (VId x t)
	return $ ForAll x $ \t -> substFormula x t fm
{- the alternative @return fm@ would not work since the translation below
   does not handle implicitely quantified variables @FOLVar@ under negation
   correctly. -}

propToFOL :: Value -> M Formula
propToFOL v =
    case vAppView v of
	VAVDef c _ []
	    | c == fol_True	-> return TRUE
	    | c == fol_False	-> return FALSE
	VAVDef c _ [v1, v2]
	    | c == fol_And	-> bin And v1 v2
	    | c == fol_Or	-> bin Or v1 v2
	    | c == fol_Implies	-> bin implies v1 v2
	    | c == fol_Equiv	-> bin equiv v1 v2
	VAVDef c _ [t, v1, v2]
	    | c == fol_Equals	->
		case unfoldConstants t of
		    -- Make sure all functions we compare are fully applied
		    VPi _ _ _	-> equal <$> mapM valueToTerm [v1',v2']
			where
			    v1' = VLam (Cl "x" $ \v -> v1 `vApp` v) t
			    v2'	= VLam (Cl "x" $ \v -> v2 `vApp` v) t
		    -- -}
		    _		-> equal <$> mapM valueToTerm [v1,v2]
	VAVDef c _ [v1]
	    | c == fol_Not	-> Not <$> propToFOL v1
	VAVDef c _ vs		-> Predicate c <$> mapM valueToTerm vs
	VAVId x _ vs		->
	    do	b <- isBoundVar x
		if b
		    then lift $ failure $ notAFOLPropositionError v
		    else Predicate x <$> mapM valueToTerm vs
	VAVPi v cl _		-> piToFOL (clName cl) v (clApp cl) propToFOL
	_   -> lift $ failure $ notAFOLPropositionError v
    where
	bin op v1 v2 = op <$> propToFOL v1 <*> propToFOL v2

valueToTerm	    = valueToTermG unfoldConstants
valueToTermNoUnfold = valueToTermG id

valueToTermG :: (Value -> Value) -> Value -> M FOLTerm
valueToTermG unfold v =
    case unfold v of
	VId x _	    ->
	    do  bound <- isBoundVar x
		return $ if bound then FOLVar x else FOLFun x []
			
	VDef x _	-> return (FOLFun x [])
	VApp v1 v2	-> app <$> mapM valToTerm [v1,v2]
	VPi w f _	->
	    do	x <- lift $ fresh (clName f)
		fun <$> mapM valToTerm [w, f `clApp` VId x w]
	VLam f t	-> lambdaLift v f
	    where VPi v _ _ = unfoldConstants t
	VTop _		-> return $ FOLFun "Top_" []
	_		-> lift $ failure $ notAFOLTermError v
    where
	valToTerm = valueToTermG unfold

