{-# OPTIONS -fglasgow-exts #-}

module Plugins where

import Prelude hiding	    ( print, putStr, putStrLn )
import IO.Output

import Control.Monad.State
import Control.Monad.Error
import Data.List		    ( isPrefixOf )
import System.Console.GetOpt	    ( OptDescr )
import System.Environment	    ( getProgName, getArgs )

import Interface.Options
import Syntax.Internal		    ( Term )
import Syntax.Values		    ( Value, Type )
import Syntax.Position		    ( Range )
import TypeChecker.Monad
import TypeChecker.Monad.Options
import TypeChecker.Monad.Errors
import TypeChecker.Errors
import Utils.Error
import Utils.Map as Map hiding (map)
import Utils.Monads

import {-# SOURCE #-} Plugins.All   ( plugins )  -- REMOVE_IF_HADDOCK

{- REMOVE_IF_HADDOCK
-- | This list is not really here. It is just an illusion to fool
--   Haddock into accepting recursive modules. The real list is
--   'Plugins.All.plugins'.
plugins :: [SomePlugin]
plugins = []
REMOVE_IF_HADDOCK -}

{- | The plugin interface. The first type parameter, @o@,is the option
record for the plugin, the second, @s@, is the plugin's internal
state. -}
data Plugin o s =
    Plugin  { plName		:: String
				{- ^ Name of the plugin, used to call
				     the plugin on a goal -}

            , plHelp            :: String
                                {- ^ Help message for this plugin,
                                     printed with --help. -}

	    , plInit		:: o -> TC s
				-- ^ Initialization

	    , plFinal		:: o -> s -> TC ()
				-- ^ Finalization (e.g. constraint solving)

	    , plCheck		:: o -> s -> Range -> [Term] -> Type -> TC (Value, s)
				-- ^ Type checking

	    , plInfer		:: o -> s -> Range -> [Term] -> TC (Value, Type, s)
				-- ^ Type inference

	    , plOptions		:: PluginOptions o
				-- ^ Command line options to the plugin

	    , plDefaultOptions	:: o
				-- ^ Default values for command line options

	    }

-- | The type checking monad carries around a map from names to plugins and
--   their states. Different plugins can have different state types, so we have
--   use existential quantification.
data SomePlugin = forall o s. SomePlugin o s (Plugin o s)

-- | Options can be inherited from other plugins.
data InheritedOptions o
	= forall o' s. InheritedOptions (Plugin o' s) (o' -> o -> o) (o -> o')

-- | Plugin options can either be both inherited from other plugins
--   and defined directly.
data PluginOptions o = PlOptions { inheritedOptions :: [InheritedOptions o]
				 , ownOptions	    :: [OptDescr (Flag o)]
				 }

simpleOptions flags = PlOptions { inheritedOptions = []
				, ownOptions	   = flags
				}

-- | Wrap a 'Plugin'. The options and initial state is undefined. They are
--   set when loading the plugin. See 'loadPlugins'.
mkPlugin :: Plugin o a -> SomePlugin
mkPlugin pl = SomePlugin undefined undefined pl

-- | Load all 'plugins'.
loadPlugins :: TC ()
loadPlugins =
    do	progName    <- liftIO getProgName
	argv	    <- liftIO getArgs
	mapM_ (loadPlugin progName argv) plugins
    where
	loadPlugin progName argv (SomePlugin _ _ pl) =
	    sayWhen (whenLoadingPlugin $ plName pl) $
		case opts' of
		    Left err	-> failure $ commandLineParseError err
		    Right opts	->
			do  interactive <- isInteractive
			    verbose 3 $ liftIO $
				putStr $ "Loading plugin "
					    ++ plName pl ++ " ..."
			    st <- plInit pl opts
			    updatePlugin (plName pl) (SomePlugin opts st pl)
			    verbose 3 $ liftIO $ putStrLn " done."
	    where
		args = [ a | a <- argv, isPluginOption pl a ]
		opts' = parsePluginOptions progName args
					   (plDefaultOptions pl)
					   (pluginOptions pl)

pluginOptions :: Plugin o s -> [OptDescr (Flag o)]
pluginOptions pl = flags ++ concatMap inh (inheritedOptions opts)
    where
	opts = plOptions pl
	inh (InheritedOptions pl inj extr) = map (fmap inject) (pluginOptions pl)
	    where
		inject flag o =
		    do	o' <- flag $ extr o
			return $ inj o' o 

	-- Prefix all the plugin flags with the name of the plugin.
	flags = map (mapFlag (\s -> plName pl ++ "-" ++ s))
		    $ ownOptions $ plOptions pl


-- | Run each plugin's finalization phase. The plugins are finalized in
--   an unspecified order (at the moment alphabetical). Maybe it would be
--   a good idea to have a more defined order?
finalizePlugins :: TC ()
finalizePlugins =
    do	pls <- tcPlugins <$> get
	mapM_ finalize $ elems pls
    where
	finalize (SomePlugin opts st pl) = plFinal pl opts st

-- | Find a particular plugin.
lookupPlugin :: String -> TC SomePlugin
lookupPlugin name =
    do	pls <- tcPlugins <$> get
	case Map.lookup name pls of
	    Nothing -> failure $ pluginNotFoundError name
	    Just pl -> return pl

-- | Update a plugin. The new plugin should only differ
--   in the state component, i.e. the actual 'Plugin'
--   should be the same.
updatePlugin :: String -> SomePlugin -> TC ()
updatePlugin name pl =
    modify $ \st -> st { tcPlugins = insert name pl $ tcPlugins st }

-- | Used when a plugin is called on a goal that is being type checked.
pluginCheck :: String -> Range -> [Term] -> Type -> TC Value
pluginCheck name r ts t =
    do	SomePlugin opts st pl <- lookupPlugin name
	(v, st') <- plCheck pl opts st r ts t
	updatePlugin name (SomePlugin opts st' pl)
	return v

-- | Used when a plugin is called on a goal whose type is being inferred.
pluginInfer :: String -> Range -> [Term] -> TC (Value, Type)
pluginInfer name r ts =
    do	SomePlugin opts st pl <- lookupPlugin name
	(v, t, st') <- plInfer pl opts st r ts
	updatePlugin name (SomePlugin opts st' pl)
	return (v, t)

-- | The names of all the plugins. Don't export.
pluginNames :: [String]
pluginNames = map name plugins
    where
	name (SomePlugin _ _ pl) = plName pl

-- | The format of a plugin specific flag. Don't export.
pluginOptionPrefix :: String -> String
pluginOptionPrefix name = "--" ++ name ++ "-"

-- | Check if a command line option does not belong to any plugin.
notPluginOption :: String -> Bool
notPluginOption s = not $ any (`isPrefixOf` s) $ map pluginOptionPrefix pluginNames

-- | Check if a command line option belongs to a particular plugin.
isPluginOption :: Plugin s o -> String -> Bool
isPluginOption pl s = any (`isPrefixOf` s)
			$ map pluginOptionPrefix
			$ plName pl : map inh (inheritedOptions $ plOptions pl)
    where
	inh (InheritedOptions pl _ _) = plName pl

-- | Only used to print the usage info. 
allPluginInfo :: [(String, String, [String], [OptDescr ()])]
allPluginInfo = map plinfo plugins
    where
	plinfo (SomePlugin _ _ pl) = 
		( plName pl
		, plHelp pl
		, map (\(InheritedOptions pl _ _) -> plName pl)
		    $ inheritedOptions $ plOptions pl
		, map (fmap (const ()) . mapFlag (\s -> plName pl ++ "-" ++ s))
			(ownOptions $ plOptions pl)
		)
