{-# OPTIONS -fglasgow-exts -fallow-overlapping-instances #-}
module TypeChecker.Patterns.Inductive
    ( PatView(..), patView
    , Tel
    , expandHidden, noWildCards
    , variablesAsMetas
    , matchPattern
    , computeBindings
--     , dropUnderscores
    ) where

import Prelude hiding	    ( print, putStr, putStrLn )
import IO.Output

import Control.Monad.Reader
import Data.Generics
import Data.IORef
import Data.List as List
import qualified Data.Set as Set

import Syntax.Position
import Syntax.Internal
import Syntax.Values

import TypeChecker.Values
import TypeChecker.Monad
import TypeChecker.Monad.Errors
import TypeChecker.Monad.Context
import TypeChecker.UnfoldConstants
import TypeChecker.Values.Refresh
import TypeChecker.Abstractions
import TypeChecker.Context
import TypeChecker.Errors
import TypeChecker.MetaVars
import TypeChecker.Monad.Definitions
import TypeChecker.Monad.Datatypes
import TypeChecker.Monad.Options
import TypeChecker.ValueToTerm

import Utils.Names
import Utils.Pairs
import Utils.Map as Map hiding (map)
import Utils.Generics
import Utils.Monads

data PatView = PVVar Name
	     | PVCon Range Name [Term] [Term]
	     | PVPair Range Term Term
	     | PVWild Range
	     | PVAbsurd Range
	     | NonPat Term
    deriving (Show)

patView :: Term -> TC PatView
patView e =
    case tAppView e of
	TAVar r x []		->
	    ifM (isBound x)
		(return $ NonPat e)
		(return $ PVVar x)
	TACon r x hs es		-> return $ PVCon r x hs es
	TAOther (Pair r e1 e2)	-> return $ PVPair r e1 e2
	TAOther (Meta r _)	-> return $ PVWild r
	TAOther (Absurd r)	-> return $ PVAbsurd r
	_			-> return $ NonPat e

variablesAsMetas :: [Term] -> TC [Term]
variablesAsMetas es = everywhereM (mkM meta) es
    where
	meta e =
	    do	pv <- patView e
		case pv of
		    PVCon _ _ _ _ -> return e
		    PVPair _ _ _  -> return e
		    PVVar x	  -> return $ Meta (getRange e) dontCare
		    _		  -> return $ Meta (getRange e) dontCare

isSoft DontCare = True
isSoft x = case nameString x of
	    '_':_   -> True
	    _	    -> False

hApp :: Term -> [Term] -> [Term] -> Term
hApp t hs es = t `happs` hs `apps` es
    where
	apps = foldl App
	happs = foldl HApp

type MetaMap = Map Name (MetaRef, Type)
type Tel = [(Name, Type)]

variableConflict r x =
    sayWhere r $ failure $ customError $ "occurences of " ++ show x ++ " conflict"

patternMismatch e v = sayWhere (getRange e) $
		      sayWhat unjustifiedPattern $
			failure $ customError
				$ "the pattern " ++ show e ++ " should be " ++ whatItShouldBe v
    where
	whatItShouldBe v =
	    case vAppView v of
		VAVMeta _ _ -> "a fresh variable"
		_	    -> show v

expandHidden :: [Term] -> TC [Term]
expandHidden = mapM expand
    where
	expand e =
	    do	pv <- patView e
		case pv of
		    PVVar x	    -> return e
		    PVCon r x hs es ->
			do  DefVal h ar _ _ _ _ <- lookupDef x
			    let xs = replicate (h - length hs)
					       (Meta r dontCare)
			    hs' <- expandHidden hs
			    es' <- expandHidden es
			    return $ hApp (Con r x) (hs' ++ xs) es'
		    PVPair r e1 e2  ->
			do  e1' <- expand e1
			    e2' <- expand e2
			    return $ Pair r e1' e2'
		    PVWild r   -> return $ Meta r dontCare
		    PVAbsurd r -> return $ Absurd r
		    NonPat e   -> return e

noWildCards :: [Term] -> TC [Term]
noWildCards = mapM noWild
    where
	 noWild e =
	    do	pv <- patView e
		case pv of
		    PVVar x	-> return e
		    PVAbsurd _	-> return e
		    NonPat e	-> return e
		    PVCon r x hs es -> hApp (Con r x) <$> noWildCards hs
						      <*> noWildCards es
		    PVPair r e1 e2	-> Pair r <$> noWild e1 <*> noWild e2
		    PVWild r	->
			do  x <- fresh "_"
			    return $ Var r x

matchPattern :: [Term] -> [(Value, Type)] -> TC ([Term], MetaMap)
matchPattern [] [] = return ([], empty)
matchPattern (e:es) ((v,t):vs) =
    do	w <- unfoldConstants v
	pv <- patView e
	case pv of
	    PVVar x	    ->
		case vAppView w of
		    VAVMeta m _  -> -- TODO: bad?
			do  (es',metaMap) <- matchPattern es vs
			    case Map.lookup x metaMap of
				Nothing	-> return (e:es', Map.insert x (m,t) metaMap)
				Just (m',_)
				    | m == m'   -> return (e:es', metaMap)
				    | otherwise ->
					variableConflict r x
		    VAVCon c _ h ar _
			| isSoft    ->
			    do	DefVal _ _ fv _ _ _ <- lookupDef c
				let h' = h - fv
				hs <- map (Var r)
				      <$> replicateM h' (fresh "_")
				xs <- map (Var r)
				      <$> replicateM (ar - h') (fresh "_")
				let e' = hApp (Con r c) hs xs
				matchPattern (e':es) ((v,t):vs)

		    _   | isSoft    ->
			    do	(es', m) <- matchPattern es vs
				return (e:es', m)

			| otherwise -> patternMismatch e w
		where
		    isSoft  = x == dontCare || take 1 (nameString x) == "_"
		    r = getRange x

	    PVCon r x hs' es'   ->
		case vAppView w of
		    VAVCon y pa h ar vs'
			| x == y    ->
			    do	DefVal _ _ fv _ _ t' <- lookupDef y
				let t'' = instantiate t' pa
				    h'	= h - fv
				verbose 18 $ liftIO $ putStrLn $ "match con " ++ show y ++ "[" ++ show fv ++ "] " ++ show vs' ++ " : " ++ show t''
				vts' <- matchPi (drop fv vs') t''
				verbose 18 $ liftIO $ putStrLn $ "  vts' = " ++ show vts'
				(es0, m) <- matchPattern (hs' ++ es' ++ es)
							 (vts' ++ vs)
				let (es1,es2) = splitAt ar es0
				    (hs,es)   = splitAt h' es1
				return (hApp (Con r x) hs es : es2, m)
			where
			    matchPi [] _ = return []
			    matchPi (v:vs) t =
				do  t' <- unfoldConstants (unEl t)
				    case vView t' of
					VPi w f	->
					    do	verbose 20 $ liftIO $ putStrLn $
						    "absApp (" ++ showVal (unEl $ absBody f) ++ ") ("
						    ++ showVal v ++ ") = " ++ showVal (unEl $ absApp f v)
						vts <- matchPi vs (absApp f v)
						return $ (v,w) : vts 
					_	-> error $ "matchPattern: should've been a Pi: " ++ show t

-- 			    nh = length $ takeWhile hidden es'
-- 			    ignore 0 h xs	= drop h xs
-- 			    ignore nh h (x:xs)	= x : ignore (nh - 1) (h - 1) xs
		    _ -> patternMismatch e w

	    PVPair r p q    ->
		case vView w of
		    VPair w1 w2	->
			do  (t1,t2) <- pairTypes w1 w2 t
			    (e1:e2:es', m) <- matchPattern (p : q : es) ((w1,t1) : (w2,t2) : vs)
			    return (Pair r e1 e2 : es',m)
			where
			    pairTypes w1 w2 t =
				do  t' <- unfoldConstants $ unEl t
				    case vView t' of
					VSigma t1 f -> return (t1, absApp f w1)
					_	    -> error $ "matchPattern: should've been a Sigma: " ++ show t
		    _	-> patternMismatch e w
	    PVWild r	    ->
		do  (es', m) <- matchPattern es vs
		    return (e:es',m)

	    PVAbsurd r	->  -- absurd patterns will be justified by the completeness checker
		do  (es', m) <- matchPattern es vs
		    return (e:es',m)

	    NonPat e	    ->
		do  justify e v
		    (es', m) <- matchPattern es vs
		    return (e:es',m)

matchPattern _ _ = error "impossible: TypeChecker.Patterns.Inductive.matchPattern"

-- data PatView = PVVar Name
-- 	     | PVCon Range Name [Term] [Term]
-- 	     | PVPair Range Term Term
-- 	     | PVWild Range
-- 	     | NonPat Term

-- | Check that the term is justified as a pattern. Amounts to checking that the
--   term and the value are the same.
justify :: Term -> Value -> TC ()
justify e v =
    do	verbose 10 $ liftIO $ putStrLn $ "justifying " ++ show e ++ " against " ++ show v
	pv <- patView e
	w  <- unfoldConstants v
	case (pv, vAppView w) of
	    (PVWild _, _)		    -> return ()
	    (PVVar _, VAVMeta _ _)	    -> return ()
	    (_, VAVMeta _ _)		    -> unjust w
	    _				    -> return ()
    where
	unjust v = sayWhere (getRange e) $ patternMismatch e v

type Renaming = Map Name Name

computeBindings :: MetaMap -> [Term] -> [Value] -> TC (Tel, [Term], [Pattern])
computeBindings metaMap es vs =
    do	(tel, es', ps, ren) <- bindings metaMap empty es vs
	verbose 6 $ liftIO $ putStrLn $ "tel  = " ++ show tel
	verbose 6 $ liftIO $ putStrLn $ "es   = " ++ show es
	verbose 6 $ liftIO $ putStrLn $ "es'  = " ++ show es'
	verbose 6 $ liftIO $ putStrLn $ "ps   = " ++ show ps
	verbose 6 $ liftIO $ putStrLn $ "ren  = " ++ show ren
	tel' <- refresh tel
	verbose 6 $ liftIO $ putStrLn $ "tel' = " ++ show tel'
	fvs <- mapM (freeVars . unEl . snd) tel'
	let tel'' = toposort $ zip tel' fvs
	verbose 6 $ liftIO $ putStrLn $ "tel'' = " ++ show tel''
	tel''' <- rebind ren tel''
	let es'' = map (rebindTerm ren) es'
	return (tel''', es'', map (rename ren) ps)
    where

	toposort [] = []
	toposort xs =
	    case emptyView xs of
		Just (((x,t),vs),xs') ->
		    (x,t) : toposort (map (del x) xs')
		Nothing	-> error $ "Panic! Cycle in type dependencies: " ++ show xs
	    where
		emptyView [] = Nothing
		emptyView (x@(_,vs):xs)
		    | Set.null vs   = Just (x,xs)
		    | otherwise	    =
			do  (y,ys) <- emptyView xs
			    return (y,x:ys)

		del v (x,vs) = (x, Set.delete v vs)

	rename ren (WildP r)	    = WildP r
	rename ren (AbsurdP r)	    = AbsurdP r
	rename ren (ConP r x ps qs) = ConP r x (map (rename ren) ps)
					       (map (rename ren) qs)
	rename ren (VarP x)	    = VarP $ findWithDefault x x ren
	rename ren (PairP r p q)    = PairP r (rename ren p) (rename ren q)

	rebind _   []		= return []
	rebind ren ((x,t):tel)  =
	    do	--liftIO $ putStrLn $ "rebinding " ++ show t
		t'   <- rebindType ren t
		let x' = findWithDefault x x ren
		--liftIO $ putStrLn $ "result: " ++ show t'
		tel' <- extendContext x' (Abstract t') $ rebind ren $ raise tel
		return ((x',t'):tel')

	rebindTerm ren e = everywhere (mkT rename) e
	    where
		rename (Var r x) = Var r $ Map.findWithDefault x x ren
		rename e	 = e

	rebindType ren t = typeMapM (rebindVal ren) t

	rebindVal ren = applyEverywhere val
	    where
		val v =
		    case vView v of
			VId x	-> fst <$> lookupVar (findWithDefault x x ren)
			_	-> return v

	freeVars :: Value -> TC (Set.Set Name)
	freeVars = foldValue Set.unions id fv
	    where
		fv v =
		    case vView v of
			VId x	-> return $ Set.singleton x
			_	-> return Set.empty

	bindings metaMap ren [] [] = return ([], [], [], ren)
	bindings metaMap ren (e:es) (v:vs) =
	    do	w <- unfoldConstants v
		verbose 10 $ liftIO $ putStrLn $ "bindings " ++ show (e:es) ++ ", " ++ show (w:vs) ++ "  " ++ show metaMap
		pv <- patView e
		case pv of
		    PVVar x	->
			case Map.lookup x metaMap of
			    Just (_,t)  ->
				case vAppView w of
				    VAVMeta m ws ->
					do  --liftIO $ putStrLn $ "  metaMap: " ++ show metaMap
					    verbose 10 $ liftIO $ putStrLn $ "  " ++ show x ++ " := " ++ show x ++ " " ++ show es ++ " " ++ show vs
					    m =: mkLams (length ws) (vId x)
					    (tel, es', ps, ren) <- bindings
								   (Map.delete x metaMap)
								   ren es vs
					    return ((x,t):tel, Var r x : es', VarP x:ps, ren)
					where
					    mkLams 0 v = v
					    mkLams n v = vLam $ mkAbs "_" $ mkLams (n - 1) v
				    VAVId y []
					| x == y || isSoft x ->
					    do  verbose 10 $ liftIO $ putStrLn $ "  " ++ show x ++ " == " ++ show w ++ " " ++ show es ++ " " ++ show vs
						wildCard (Var r y)
					| isSoft y  ->
					    case Map.lookup y ren of
						Nothing	->
						    do	verbose 10 $ liftIO $ putStrLn $ "  renaming " ++ show y ++ " to " ++ show x
							wildCard' (Map.insert y x ren) (Var r x)
						Just z
						    | x == z	-> wildCard (Var r x)
						    | otherwise	-> patternMismatch e w
				    _	-> patternMismatch e w
			    Nothing
				| isSoft x  -> wildCard =<< valueToTerm' meta w
				| otherwise -> wildCard (Var r x)
				where
				    metas = map (fst . snd /\ fst) $ Map.toList metaMap
				    meta m _ = case List.lookup m metas of
						 Just y  -> return $ Var r y
						 Nothing ->
						    sayWhere r
						    $ failure $ hiddenArgumentInSoftPatternError e v
			where
			    r = getRange x

		    PVCon r x hs' es'	->
			case vAppView w of
			    VAVCon y pa h ar vs' ->
				do  dv@(DefVal _ _ fv _ _ _) <- lookupDef y
				    verbose 12 $ liftIO $ putStrLn $ "binding con: " ++ show dv
				    (tel, es, ps, ren) <- bindings metaMap ren
							  (hs' ++ es' ++ es)
							  (drop fv vs' ++ vs)
				    let h' = h - fv
				    unless (length hs' == h' && length es' == ar - h')
					$ error $ "TypeChecker.Patterns.Inductive.bindings: Panic: " ++
						    unwords ["length", show hs', "!=", show h'
							    , "||", "length", show es', "!=", show ar, "-", show h']
				    let (ps0,ps1) = splitAt ar ps
					(hs,ps0') = splitAt h' ps0
					(es0,es1) = splitAt ar es
					(ehs,es0')= splitAt h' es0
				    return ( tel, hApp (Con r x) ehs es0' : es1
					   , ConP r x hs ps0' : ps1, ren
					   )
			    _	-> error "impossible: TypeChecker.Patterns.Inductive.bindings (PVCon _ _ _ _)"

		    PVPair r e1 e2	->
			case vView w of
			    VPair v1 v2 ->
				do  (tel, e1:e2:es, p:q:ps, ren) <-
					    bindings metaMap ren
					    (e1 : e2 : es)
					    (v1 : v2 : vs)
				    return (tel, Pair r e1 e2 : es, PairP r p q : ps, ren)
			    _	-> error "impossible: TypeChecker.Patterns.Inductive.bindings (PVPair _ _ _)"
		    PVWild r	-> failure $ customError $ "didn't we get rid of these? " ++ show (e:es) ++ " " ++ show (v:vs)
				    --wildCard
		    PVAbsurd r	->
			do  (tel, es', ps, ren) <- bindings metaMap ren es vs
			    return (tel, e:es', AbsurdP (getRange e) : ps, ren)

		    NonPat e'	-> wildCard e'
	    where
		wildCard = wildCard' ren
		wildCard' ren e =
		    do	(tel, es', ps, ren) <- bindings metaMap ren es vs
			return (tel, e:es', WildP (getRange e) : ps, ren)
	bindings _ _ es vs = failure $ customError $ "impossible: TypeChecker.Patterns.Inductive.bindings " ++ show es ++ " " ++ show vs

